@@ -11,10 +11,10 @@ pub enum Any {
     Bool(bool),
     Number(f64),
     BigInt(i64),
-    String(String),
+    String(Box<str>),
     Buffer(Box<[u8]>),
-    Array(Vec<Any>),
-    Map(HashMap<String, Any>),
+    Array(Box<[Any]>),
+    Map(Box<HashMap<String, Any>>),
 }
 
 impl Any {
@@ -37,7 +37,7 @@ impl Any {
             // CASE 120: boolean (true)
             120 => Any::Bool(true),
             // CASE 119: string
-            119 => Any::String(decoder.read_string().to_owned()),
+            119 => Any::String(decoder.read_string().into()),
             // CASE 118: Map<string,Any>
             118 => {
                 let len: usize = decoder.read_uvar();
@@ -46,7 +46,7 @@ impl Any {
                     let key = decoder.read_string();
                     map.insert(key.to_owned(), Any::decode(decoder));
                 }
-                Any::Map(map)
+                Any::Map(Box::new(map))
             }
             // CASE 117: Array<Any>
             117 => {
@@ -55,7 +55,7 @@ impl Any {
                 for _ in 0..len {
                     arr.push(Any::decode(decoder));
                 }
-                Any::Array(arr)
+                Any::Array(arr.into_boxed_slice())
             }
             // CASE 116: buffer
             116 => Any::Buffer(Box::from(decoder.read_buf().to_owned())),
@@ -152,7 +152,7 @@ impl Any {
                 // TYPE 118: Map
                 encoder.write_u8(118);
                 encoder.write_uvar(map.len() as u64);
-                for (key, value) in map {
+                for (key, value) in map.as_ref() {
                     encoder.write_string(&key);
                     value.encode(encoder);
                 }
@@ -164,18 +164,86 @@ impl Any {
             }
         }
     }
+
+    pub fn to_json(&self, buf: &mut String) {
+        use std::fmt::Write;
+
+        fn quoted(buf: &mut String, s: &str) {
+            buf.reserve(s.len() + 2);
+            buf.push('"');
+            for c in s.chars() {
+                match c {
+                    '\\' => buf.push_str("\\\\"),
+                    '\u{0008}' => buf.push_str("\\b"),
+                    '\u{000c}' => buf.push_str("\\f"),
+                    '\n' => buf.push_str("\\n"),
+                    '\r' => buf.push_str("\\r"),
+                    '\t' => buf.push_str("\\t"),
+                    '"' => buf.push_str("\\\""),
+                    c if c.is_control() => write!(buf, "\\u{:04x}", c as u32).unwrap(),
+                    c => buf.push(c),
+                }
+            }
+            buf.push('"');
+        }
+
+        match self {
+            Any::Null => buf.push_str("null"),
+            Any::Bool(value) => write!(buf, "{}", value).unwrap(),
+            Any::Number(value) => write!(buf, "{}", value).unwrap(),
+            Any::BigInt(value) => write!(buf, "{}", value).unwrap(),
+            Any::String(value) => quoted(buf, value.as_ref()),
+            Any::Array(values) => {
+                buf.push('[');
+                let mut i = values.iter();
+                if let Some(value) = i.next() {
+                    value.to_json(buf);
+                }
+                while let Some(value) = i.next() {
+                    buf.push(',');
+                    value.to_json(buf);
+                }
+                buf.push(']');
+            }
+            Any::Map(entries) => {
+                buf.push('{');
+                let mut i = entries.iter();
+                if let Some((key, value)) = i.next() {
+                    quoted(buf, key.as_str());
+                    buf.push(':');
+                    value.to_json(buf);
+                }
+                while let Some((key, value)) = i.next() {
+                    buf.push(',');
+                    quoted(buf, key.as_str());
+                    buf.push(':');
+                    value.to_json(buf);
+                }
+                buf.push('}');
+            }
+            other => panic!(
+                "Serialization of {} into JSON representation is not supported",
+                other
+            ),
+        }
+    }
+
+    pub fn from_json(src: &str) -> Self {
+        use crate::json_parser::JsonParser;
+        let mut parser = JsonParser::new(src.chars());
+        parser.parse().unwrap()
+    }
 }
 
 impl std::fmt::Display for Any {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         match self {
-            Any::Null => write!(f, "null")?,
-            Any::Undefined => write!(f, "undefined")?,
-            Any::Bool(value) => write!(f, "{}", value)?,
-            Any::Number(value) => write!(f, "{}", value)?,
-            Any::BigInt(value) => write!(f, "{}", value)?,
-            Any::String(value) => write!(f, "{}", value)?,
-            Any::Buffer(value) => write!(f, "[binary: {} bytes]", value.len())?, //TODO: use base64?
+            Any::Null => f.write_str("null"),
+            Any::Undefined => f.write_str("undefined"),
+            Any::Bool(value) => write!(f, "{}", value),
+            Any::Number(value) => write!(f, "{}", value),
+            Any::BigInt(value) => write!(f, "{}", value),
+            Any::String(value) => f.write_str(value.as_ref()),
             Any::Array(values) => {
                 write!(f, "[")?;
                 let mut i = values.iter();
@@ -186,22 +254,27 @@ impl std::fmt::Display for Any {
                     write!(f, ", ")?;
                     value.fmt(f)?;
                 }
-                write!(f, "]")?;
+                write!(f, "]")
             }
             Any::Map(entries) => {
                 write!(f, "{{")?;
                 let mut i = entries.iter();
                 if let Some((key, value)) = i.next() {
-                    write!(f, "\"{}\": {}", key, value)?;
+                    write!(f, "{}: {}", key, value)?;
                 }
                 while let Some((key, value)) = i.next() {
-                    write!(f, ", \"{}\": {}", key, value)?;
+                    write!(f, ", {}: {}", key, value)?;
                 }
-                write!(f, "}}")?;
+                write!(f, "}}")
+            }
+            Any::Buffer(value) => {
+                f.write_str("0x")?;
+                for &byte in value.iter() {
+                    write!(f, "{:02x}", byte)?;
+                }
+                Ok(())
             }
         }
-
-        Ok(())
     }
 }
 
@@ -237,13 +310,13 @@ impl Into<Any> for i32 {
 
 impl Into<Any> for String {
     fn into(self) -> Any {
-        Any::String(self)
+        Any::String(self.into_boxed_str())
     }
 }
 
 impl Into<Any> for &str {
     fn into(self) -> Any {
-        Any::String(self.to_string())
+        Any::String(self.into())
     }
 }
 
@@ -280,7 +353,7 @@ where
         for value in self {
             array.push(value.into())
         }
-        Any::Array(array)
+        Any::Array(array.into_boxed_slice())
     }
 }
 
@@ -293,7 +366,7 @@ where
         for (key, value) in self {
             map.insert(key, value.into());
         }
-        Any::Map(map)
+        Any::Map(Box::new(map))
     }
 }
 
