@@ -9,6 +9,7 @@ use std::collections::hash_map::Entry;
 use std::collections::HashMap;
 use std::hash::BuildHasherDefault;
 use std::ops::Index;
+use std::rc::Rc;
 use std::vec::Vec;
 
 /// State vector is a compact representation of all known blocks inserted and integrated into
@@ -133,6 +134,40 @@ impl Encode for StateVector {
     }
 }
 
+#[derive(Default, Debug, Clone, PartialEq, Eq)]
+pub struct Snapshot {
+    pub delete_set: DeleteSet,
+    pub state_map: StateVector,
+}
+
+impl Snapshot {
+    pub fn new(state_map: StateVector, delete_set: DeleteSet) -> Self {
+        Snapshot {
+            state_map,
+            delete_set,
+        }
+    }
+
+    pub(crate) fn is_visible(&self, id: &ID) -> bool {
+        self.state_map.contains(id) && !self.delete_set.is_deleted(id)
+    }
+}
+
+impl Encode for Snapshot {
+    fn encode<E: Encoder>(&self, encoder: &mut E) {
+        self.delete_set.encode(encoder);
+        self.state_map.encode(encoder);
+    }
+}
+
+impl Decode for Snapshot {
+    fn decode<D: Decoder>(decoder: &mut D) -> Self {
+        let ds = DeleteSet::decode(decoder);
+        let sm = StateVector::decode(decoder);
+        Snapshot::new(sm, ds)
+    }
+}
+
 /// A resizable list of blocks inserted by a single client.
 pub(crate) struct ClientBlockList {
     list: Vec<UnsafeCell<block::Block>>,
@@ -354,7 +389,7 @@ impl ClientBlockList {
 #[derive(Debug)]
 pub(crate) struct SquashResult {
     pub parent: TypePtr,
-    pub parent_sub: Option<String>,
+    pub parent_sub: Option<Rc<str>>,
     /// Pointer to a block that resulted from compaction of two adjacent blocks.
     pub replacement: BlockPtr,
     /// Pointer to a neighbor, that's now on the right side of the `replacement` block.
@@ -419,11 +454,6 @@ impl BlockStore {
         self.clients.is_empty()
     }
 
-    /// Checks if the were any blocks integrated from a given `client`.
-    pub fn contains_client(&self, client: &u64) -> bool {
-        self.clients.contains_key(client)
-    }
-
     /// Returns an immutable reference to a block list for a particular `client`. Returns `None` if
     /// no block list existed for provided `client` in current block store.
     pub fn get(&self, client: &u64) -> Option<&ClientBlockList> {
@@ -566,16 +596,26 @@ impl BlockStore {
                         let right_split_id = right_split.id.clone();
                         let right_ptr = right_split.right.clone();
                         if let Some(right_ptr) = right_ptr {
-                            blocks = if right_ptr.id.client == ptr.id.client {
-                                blocks
+                            let right_left =
+                                Some(BlockPtr::new(right_split.id.clone(), index as u32));
+
+                            if right_ptr.id.client == ptr.id.client {
+                                if let Block::Item(item) = blocks.find(&right_ptr).unwrap() {
+                                    item.left = right_left;
+                                }
                             } else {
-                                self.clients.get_mut(&right_ptr.id.client).unwrap()
+                                if let Block::Item(item) = self
+                                    .clients
+                                    .get_mut(&right_ptr.id.client)
+                                    .unwrap()
+                                    .find(&right_ptr)
+                                    .unwrap()
+                                {
+                                    item.left = right_left;
+                                }
+
+                                blocks = self.clients.get_mut(&ptr.id.client).unwrap();
                             };
-                            let right = blocks.find(&right_ptr).unwrap();
-                            if let Some(right_item) = right.as_item_mut() {
-                                right_item.left =
-                                    Some(BlockPtr::new(right_split.id.clone(), index as u32));
-                            }
                         }
                         blocks.insert(index, Block::Item(right_split));
                         Some(BlockPtr::new(right_split_id, index as u32))
