@@ -43,7 +43,7 @@ fn text_insert_delete() {
             None,
             None,
             None,
-            TypePtr::Named(Rc::new("type".to_string())),
+            TypePtr::Named("type".into()),
             None,
             ItemContent::Deleted(3),
         )),
@@ -55,7 +55,7 @@ fn text_insert_delete() {
             Some(ID::new(CLIENT_ID, 0)),
             TypePtr::Unknown,
             None,
-            ItemContent::String("ab".to_string()),
+            ItemContent::String("ab".into()),
         )),
         Block::Item(Item::new(
             ID::new(CLIENT_ID, 5),
@@ -85,7 +85,7 @@ fn text_insert_delete() {
             None,
             TypePtr::Unknown,
             None,
-            ItemContent::String("hi".to_string()),
+            ItemContent::String("hi".into()),
         )),
     ];
     let expected_ds = {
@@ -98,7 +98,7 @@ fn text_insert_delete() {
     let setter = visited.clone();
 
     let mut doc = Doc::new();
-    let _sub = doc.on_update(move |e| {
+    let _sub = doc.on_update(move |_, e| {
         for (actual, expected) in e.update.blocks.blocks().zip(expected_blocks.as_slice()) {
             assert_eq!(actual, expected);
         }
@@ -136,9 +136,9 @@ fn map_set() {
             None,
             None,
             None,
-            TypePtr::Named(Rc::new("test".to_string())),
-            Some("k1".to_string()),
-            ItemContent::Any(vec![Any::String("v1".to_string())]),
+            TypePtr::Named("test".into()),
+            Some("k1".into()),
+            ItemContent::Any(vec![Any::String("v1".into())]),
         )),
         &Block::Item(Item::new(
             ID::new(CLIENT_ID, 1),
@@ -146,9 +146,9 @@ fn map_set() {
             None,
             None,
             None,
-            TypePtr::Named(Rc::new("test".to_string())),
-            Some("k2".to_string()),
-            ItemContent::Any(vec![Any::String("v2".to_string())]),
+            TypePtr::Named("test".into()),
+            Some("k2".into()),
+            ItemContent::Any(vec![Any::String("v2".into())]),
         )),
     ];
 
@@ -177,12 +177,9 @@ fn array_insert() {
         None,
         None,
         None,
-        TypePtr::Named(Rc::new("test".to_string())),
+        TypePtr::Named("test".into()),
         None,
-        ItemContent::Any(vec![
-            Any::String("a".to_string()),
-            Any::String("b".to_string()),
-        ]),
+        ItemContent::Any(vec![Any::String("a".into()), Any::String("b".into())]),
     ))];
 
     roundtrip(payload, expected);
@@ -214,7 +211,7 @@ fn xml_fragment_insert() {
             None,
             None,
             None,
-            TypePtr::Named(Rc::new("fragment-name".to_string())),
+            TypePtr::Named("fragment-name".into()),
             None,
             ItemContent::Type(BranchRef::new(Branch::new(
                 TypePtr::Id(BlockPtr::from(ID::new(CLIENT_ID, 0))),
