@@ -1,8 +1,13 @@
 use crate::block::{BlockPtr, ItemContent, ItemPosition, Prelim};
-use crate::types::{Branch, BranchRef, TypePtr, Value, TYPE_REFS_ARRAY};
-use crate::Transaction;
+use crate::event::Subscription;
+use crate::types::{
+    event_change_set, Branch, BranchRef, Change, ChangeSet, Observers, Path, TypePtr, Value,
+    TYPE_REFS_ARRAY,
+};
+use crate::{SubscriptionId, Transaction, ID};
 use lib0::any::Any;
-use std::collections::VecDeque;
+use std::cell::UnsafeCell;
+use std::collections::{HashSet, VecDeque};
 
 /// A collection used to store data in an indexed sequence structure. This type is internally
 /// implemented as a double linked list, which may squash values inserted directly one after another
@@ -56,6 +61,7 @@ impl Array {
             left,
             right,
             index: 0,
+            current_attrs: None,
         };
 
         txn.create_item(&pos, value, None);
@@ -105,13 +111,13 @@ impl Array {
     /// of the range of a current array.
     pub fn get(&self, txn: &Transaction, index: u32) -> Option<Value> {
         let inner = self.0.borrow();
-        let (content, idx) = inner.get_at(txn, index)?;
+        let (content, idx) = inner.get_at(&txn.store().blocks, index)?;
         Some(content.get_content(txn).remove(idx))
     }
 
     /// Returns an iterator, that can be used to lazely traverse over all values stored in a current
     /// array.
-    pub fn iter<'a, 'b, 'txn>(&'a self, txn: &'b Transaction<'txn>) -> ArrayIter<'b, 'txn> {
+    pub fn iter<'a, 'b>(&'a self, txn: &'b Transaction) -> ArrayIter<'b> {
         ArrayIter::new(self, txn)
     }
 
@@ -120,16 +126,43 @@ impl Array {
         let res = self.iter(txn).map(|v| v.to_json(txn)).collect();
         Any::Array(res)
     }
+
+    /// Subscribes a given callback to be triggered whenever current array is changed.
+    /// A callback is triggered whenever a transaction gets committed. This function does not
+    /// trigger if changes have been observed by nested shared collections.
+    ///
+    /// All array changes can be tracked by using [Event::delta] method.
+    ///
+    /// Returns an [Observer] which, when dropped, will unsubscribe current callback.
+    pub fn observe<F>(&self, f: F) -> Subscription<ArrayEvent>
+    where
+        F: Fn(&Transaction, &ArrayEvent) -> () + 'static,
+    {
+        let mut branch = self.0.borrow_mut();
+        if let Observers::Array(eh) = branch.observers.get_or_insert_with(Observers::array) {
+            eh.subscribe(f)
+        } else {
+            panic!("Observed collection is of different type") //TODO: this should be Result::Err
+        }
+    }
+
+    /// Unsubscribes a previously subscribed event callback identified by given `subscription_id`.
+    pub fn unobserve(&self, subscription_id: SubscriptionId) {
+        let mut branch = self.0.borrow_mut();
+        if let Some(Observers::Array(eh)) = branch.observers.as_mut() {
+            eh.unsubscribe(subscription_id);
+        }
+    }
 }
 
-pub struct ArrayIter<'b, 'txn> {
+pub struct ArrayIter<'b> {
     content: VecDeque<Value>,
     ptr: Option<BlockPtr>,
-    txn: &'b Transaction<'txn>,
+    txn: &'b Transaction,
 }
 
-impl<'b, 'txn> ArrayIter<'b, 'txn> {
-    fn new(array: &Array, txn: &'b Transaction<'txn>) -> Self {
+impl<'b> ArrayIter<'b> {
+    fn new(array: &Array, txn: &'b Transaction) -> Self {
         let inner = array.0.borrow();
         ArrayIter {
             ptr: inner.start,
@@ -139,14 +172,14 @@ impl<'b, 'txn> ArrayIter<'b, 'txn> {
     }
 }
 
-impl<'b, 'txn> Iterator for ArrayIter<'b, 'txn> {
+impl<'b> Iterator for ArrayIter<'b> {
     type Item = Value;
 
     fn next(&mut self) -> Option<Self::Item> {
         match self.content.pop_front() {
             None => {
                 if let Some(ptr) = self.ptr.take() {
-                    let item = self.txn.store.blocks.get_item(&ptr)?;
+                    let item = self.txn.store().blocks.get_item(&ptr)?;
                     self.ptr = item.right.clone();
                     if !item.is_deleted() && item.is_countable() {
                         self.content = item.content.get_content(self.txn).into();
@@ -220,16 +253,71 @@ where
     }
 }
 
+/// Event generated by [Array::observe] method. Emitted during transaction commit phase.
+pub struct ArrayEvent {
+    target: Array,
+    current_target: BranchRef,
+    change_set: UnsafeCell<Option<Box<ChangeSet<Change>>>>,
+}
+
+impl ArrayEvent {
+    pub(crate) fn new(branch_ref: BranchRef) -> Self {
+        let current_target = branch_ref.clone();
+        ArrayEvent {
+            target: Array::from(branch_ref),
+            current_target,
+            change_set: UnsafeCell::new(None),
+        }
+    }
+
+    /// Returns an [Array] instance which emitted this event.
+    pub fn target(&self) -> &Array {
+        &self.target
+    }
+
+    /// Returns a path from root type down to [Text] instance which emitted this event.
+    pub fn path(&self, txn: &Transaction) -> Path {
+        Branch::path(self.current_target.borrow(), self.target.0.borrow(), txn)
+    }
+
+    /// Returns summary of changes made over corresponding [Array] collection within
+    /// a bounds of current transaction.
+    pub fn delta(&self, txn: &Transaction) -> &[Change] {
+        self.changes(txn).delta.as_slice()
+    }
+
+    /// Returns a collection of block identifiers that have been added within a bounds of
+    /// current transaction.
+    pub fn inserts(&self, txn: &Transaction) -> &HashSet<ID> {
+        &self.changes(txn).added
+    }
+
+    /// Returns a collection of block identifiers that have been removed within a bounds of
+    /// current transaction.
+    pub fn removes(&self, txn: &Transaction) -> &HashSet<ID> {
+        &self.changes(txn).deleted
+    }
+
+    fn changes(&self, txn: &Transaction) -> &ChangeSet<Change> {
+        let change_set = unsafe { self.change_set.get().as_mut().unwrap() };
+        change_set.get_or_insert_with(|| {
+            Box::new(event_change_set(txn, self.target.0.borrow().start.as_ref()))
+        })
+    }
+}
+
 #[cfg(test)]
 mod test {
     use crate::test_utils::{exchange_updates, run_scenario, RngExt};
     use crate::types::map::PrelimMap;
-    use crate::types::Value;
-    use crate::{Doc, PrelimArray};
+    use crate::types::{Change, Value};
+    use crate::{Doc, PrelimArray, Update, ID};
     use lib0::any::Any;
     use rand::prelude::StdRng;
     use rand::Rng;
-    use std::collections::HashMap;
+    use std::cell::{Cell, RefCell};
+    use std::collections::{HashMap, HashSet};
+    use std::rc::Rc;
 
     #[test]
     fn push_back() {
@@ -567,13 +655,194 @@ mod test {
             expected.insert("value".to_owned(), Any::Number(i as f64));
             match value {
                 Value::YMap(_) => {
-                    assert_eq!(value.to_json(&txn), Any::Map(expected))
+                    assert_eq!(value.to_json(&txn), Any::Map(Box::new(expected)))
                 }
                 _ => panic!("Value of array at index {} was no YMap", i),
             }
         }
     }
 
+    #[test]
+    fn insert_and_remove_events() {
+        let d = Doc::with_client_id(1);
+        let array = {
+            let mut txn = d.transact();
+            txn.get_array("array")
+        };
+        let happened = Rc::new(Cell::new(false));
+        let happened_clone = happened.clone();
+        let _sub = array.observe(move |_, _| {
+            happened_clone.set(true);
+        });
+
+        {
+            let mut txn = d.transact();
+            array.insert_range(&mut txn, 0, [0, 1, 2]);
+            // txn is committed at the end of this scope
+        }
+        assert!(
+            happened.replace(false),
+            "insert of [0,1,2] should trigger event"
+        );
+
+        {
+            let mut txn = d.transact();
+            array.remove_range(&mut txn, 0, 1);
+            // txn is committed at the end of this scope
+        }
+        assert!(
+            happened.replace(false),
+            "removal of [0] should trigger event"
+        );
+
+        {
+            let mut txn = d.transact();
+            array.remove_range(&mut txn, 0, 2);
+            // txn is committed at the end of this scope
+        }
+        assert!(
+            happened.replace(false),
+            "removal of [1,2] should trigger event"
+        );
+    }
+
+    #[test]
+    fn insert_and_remove_event_changes() {
+        let d1 = Doc::with_client_id(1);
+        let array = {
+            let mut txn = d1.transact();
+            txn.get_array("array")
+        };
+        let added = Rc::new(RefCell::new(None));
+        let removed = Rc::new(RefCell::new(None));
+        let delta = Rc::new(RefCell::new(None));
+
+        let (added_c, removed_c, delta_c) = (added.clone(), removed.clone(), delta.clone());
+        let _sub = array.observe(move |txn, e| {
+            *added_c.borrow_mut() = Some(e.inserts(txn).clone());
+            *removed_c.borrow_mut() = Some(e.removes(txn).clone());
+            *delta_c.borrow_mut() = Some(e.delta(txn).to_vec());
+        });
+
+        {
+            let mut txn = d1.transact();
+            array.push_back(&mut txn, 4);
+            array.push_back(&mut txn, "dtrn");
+            // txn is committed at the end of this scope
+        }
+        assert_eq!(
+            added.borrow_mut().take(),
+            Some(HashSet::from([ID::new(1, 0), ID::new(1, 1)]))
+        );
+        assert_eq!(removed.borrow_mut().take(), Some(HashSet::new()));
+        assert_eq!(
+            delta.borrow_mut().take(),
+            Some(vec![Change::Added(vec![
+                Any::Number(4.0).into(),
+                Any::String("dtrn".into()).into()
+            ])])
+        );
+
+        {
+            let mut txn = d1.transact();
+            array.remove_range(&mut txn, 0, 1);
+        }
+        assert_eq!(added.borrow_mut().take(), Some(HashSet::new()));
+        assert_eq!(
+            removed.borrow_mut().take(),
+            Some(HashSet::from([ID::new(1, 0)]))
+        );
+        assert_eq!(delta.borrow_mut().take(), Some(vec![Change::Removed(1)]));
+
+        {
+            let mut txn = d1.transact();
+            array.insert(&mut txn, 1, 0.5);
+        }
+        assert_eq!(
+            added.borrow_mut().take(),
+            Some(HashSet::from([ID::new(1, 2)]))
+        );
+        assert_eq!(removed.borrow_mut().take(), Some(HashSet::new()));
+        assert_eq!(
+            delta.borrow_mut().take(),
+            Some(vec![
+                Change::Retain(1),
+                Change::Added(vec![Any::Number(0.5).into()])
+            ])
+        );
+
+        let d2 = Doc::with_client_id(2);
+        let array2 = {
+            let mut txn = d2.transact();
+            txn.get_array("array")
+        };
+        let (added_c, removed_c, delta_c) = (added.clone(), removed.clone(), delta.clone());
+        let _sub = array2.observe(move |txn, e| {
+            *added_c.borrow_mut() = Some(e.inserts(txn).clone());
+            *removed_c.borrow_mut() = Some(e.removes(txn).clone());
+            *delta_c.borrow_mut() = Some(e.delta(txn).to_vec());
+        });
+
+        {
+            let t1 = d1.transact();
+            let mut t2 = d2.transact();
+
+            let sv = t2.state_vector();
+            let mut encoder = EncoderV1::new();
+            t1.encode_diff(&sv, &mut encoder);
+            t2.apply_update(Update::decode_v1(encoder.to_vec().as_slice()));
+        }
+
+        assert_eq!(
+            added.borrow_mut().take(),
+            Some(HashSet::from([ID::new(1, 1)]))
+        );
+        assert_eq!(removed.borrow_mut().take(), Some(HashSet::new()));
+        assert_eq!(
+            delta.borrow_mut().take(),
+            Some(vec![Change::Added(vec![
+                Any::String("dtrn".into()).into(),
+                Any::Number(0.5).into(),
+            ])])
+        );
+    }
+
+    #[test]
+    fn target_on_local_and_remote() {
+        let d1 = Doc::with_client_id(1);
+        let d2 = Doc::with_client_id(2);
+        let a1 = {
+            let mut txn = d1.transact();
+            txn.get_array("array")
+        };
+        let a2 = {
+            let mut txn = d2.transact();
+            txn.get_array("array")
+        };
+
+        let c1 = Rc::new(RefCell::new(None));
+        let c1c = c1.clone();
+        let _s1 = a1.observe(move |_, e| {
+            *c1c.borrow_mut() = Some(e.target().clone());
+        });
+        let c2 = Rc::new(RefCell::new(None));
+        let c2c = c2.clone();
+        let _s2 = a2.observe(move |_, e| {
+            *c2c.borrow_mut() = Some(e.target().clone());
+        });
+
+        {
+            let mut t1 = d1.transact();
+            a1.insert_range(&mut t1, 0, [1, 2]);
+        }
+        exchange_updates(&[&d1, &d2]);
+
+        assert_eq!(c1.borrow_mut().take(), Some(a1));
+        assert_eq!(c2.borrow_mut().take(), Some(a2));
+    }
+
+    use crate::updates::decoder::Decode;
+    use crate::updates::encoder::{Encoder, EncoderV1};
     use std::sync::atomic::{AtomicI64, Ordering};
 
     static UNIQUE_NUMBER: AtomicI64 = AtomicI64::new(0);
@@ -593,7 +862,8 @@ mod test {
                 .map(|_| Any::BigInt(unique_number))
                 .collect();
             let mut pos = rng.between(0, yarray.len()) as usize;
-            if let Any::Array(mut expected) = yarray.to_json(&txn) {
+            if let Any::Array(expected) = yarray.to_json(&txn) {
+                let mut expected = Vec::from(expected);
                 yarray.insert_range(&mut txn, pos as u32, content.clone());
 
                 for any in content {
@@ -601,7 +871,7 @@ mod test {
                     pos += 1;
                 }
                 let actual = yarray.to_json(&txn);
-                assert_eq!(actual, Any::Array(expected))
+                assert_eq!(actual, Any::Array(expected.into_boxed_slice()))
             } else {
                 panic!("should not happen")
             }
@@ -613,7 +883,7 @@ mod test {
             let pos = rng.between(0, yarray.len());
             yarray.insert(&mut txn, pos, PrelimArray::from([1, 2, 3, 4]));
             if let Value::YArray(array2) = yarray.get(&txn, pos).unwrap() {
-                let expected: Vec<_> = (1..=4).map(|i| Any::Number(i as f64)).collect();
+                let expected: Box<[Any]> = (1..=4).map(|i| Any::Number(i as f64)).collect();
                 assert_eq!(array2.to_json(&txn), Any::Array(expected));
             } else {
                 panic!("should not happen")
@@ -630,7 +900,7 @@ mod test {
                 map.insert(&mut txn, "someprop".to_string(), 43);
                 map.insert(&mut txn, "someprop".to_string(), 44);
             } else {
-                panic!("should not happen: {}", txn.store)
+                panic!("should not happen: {}", txn.store())
             }
         }
 
@@ -648,10 +918,14 @@ mod test {
                         array2.remove_range(&mut txn, pos, del_len);
                     }
                 } else {
-                    if let Any::Array(mut old_content) = yarray.to_json(&txn) {
+                    if let Any::Array(old_content) = yarray.to_json(&txn) {
+                        let mut old_content = Vec::from(old_content);
                         yarray.remove_range(&mut txn, pos, del_len);
                         old_content.drain(pos as usize..(pos + del_len) as usize);
-                        assert_eq!(yarray.to_json(&txn), Any::Array(old_content));
+                        assert_eq!(
+                            yarray.to_json(&txn),
+                            Any::Array(old_content.into_boxed_slice())
+                        );
                     } else {
                         panic!("should not happen")
                     }
