@@ -8,11 +8,15 @@ pub use map::Map;
 pub use text::Text;
 
 use crate::block::{BlockPtr, Item, ItemContent, ItemPosition, Prelim};
-use crate::types::array::Array;
-use crate::types::xml::{XmlElement, XmlText};
+use crate::block_store::BlockStore;
+use crate::event::EventHandler;
+use crate::types::array::{Array, ArrayEvent};
+use crate::types::map::MapEvent;
+use crate::types::text::TextEvent;
+use crate::types::xml::{XmlElement, XmlEvent, XmlText, XmlTextEvent};
 use lib0::any::Any;
 use std::cell::{BorrowMutError, Ref, RefCell, RefMut};
-use std::collections::HashMap;
+use std::collections::{HashMap, HashSet, VecDeque};
 use std::fmt::Formatter;
 use std::rc::Rc;
 
@@ -105,15 +109,21 @@ impl BranchRef {
         };
         while remaining > 0 {
             if let Some(mut p) = ptr {
-                if let Some(item) = txn.store.blocks.get_item(&p) {
+                let encoding = txn.store().options.offset_kind;
+                if let Some(item) = txn.store().blocks.get_item(&p) {
                     if !item.is_deleted() {
-                        let item_len = item.len();
-                        let (l, r) = if remaining < item_len {
-                            p.id.clock += remaining;
+                        let content_len = item.content_len(encoding);
+                        let (l, r) = if remaining < content_len {
+                            let offset = if let ItemContent::String(s) = &item.content {
+                                s.block_offset(remaining, encoding)
+                            } else {
+                                remaining
+                            };
+                            p.id.clock += offset;
                             remaining = 0;
-                            txn.store.blocks.split_block(&p)
+                            txn.store_mut().blocks.split_block(&p)
                         } else {
-                            remaining -= item_len;
+                            remaining -= content_len;
                             (ptr, item.right.clone())
                         };
                         txn.delete(&l.unwrap());
@@ -156,10 +166,17 @@ impl BranchRef {
             left,
             right,
             index: 0,
+            current_attrs: None,
         };
 
         txn.create_item(&pos, value, None)
     }
+
+    pub(crate) fn trigger(&self, txn: &Transaction, keys: HashSet<Option<Rc<str>>>) {
+        if let Some(o) = self.borrow().observers.as_ref() {
+            o.publish(self.clone(), txn, keys);
+        }
+    }
 }
 
 impl AsRef<Branch> for BranchRef {
@@ -189,7 +206,6 @@ impl PartialEq for BranchRef {
 }
 
 /// Branch describes a content of a complex Yrs data structures, such as arrays or maps.
-#[derive(Debug, Clone, Eq, PartialEq)]
 pub struct Branch {
     /// A pointer to a first block of a indexed sequence component of this branch node. If `None`,
     /// it means that sequence is empty or a branch doesn't act as an indexed sequence. Indexed
@@ -209,7 +225,7 @@ pub struct Branch {
     /// - [Map]: all of the map elements are based on this field. The value of each entry points
     ///   to the last modified value.
     /// - [XmlElement]: this field stores attributes assigned to a given XML node.
-    pub map: HashMap<String, BlockPtr>,
+    pub map: HashMap<Rc<str>, BlockPtr>,
 
     /// Unique identifier of a current branch node. It can be contain either a named string - which
     /// means, this branch is a root-level complex data structure - or a block identifier. In latter
@@ -220,14 +236,42 @@ pub struct Branch {
     /// A tag name identifier, used only by [XmlElement].
     pub name: Option<String>,
 
-    pub item: Option<BlockPtr>, //TODO: isn't this equivalent to `ptr` field?
-
     /// A length of an indexed sequence component of a current branch node. Map component elements
     /// are computed on demand.
-    pub len: u32,
+    pub block_len: u32,
+
+    pub content_len: u32,
 
     /// An identifier of an underlying complex data type (eg. is it an Array or a Map).
     type_ref: TypeRefs,
+
+    observers: Option<Observers>,
+}
+
+impl std::fmt::Debug for Branch {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        f.debug_struct("Branch")
+            .field("start", &self.start)
+            .field("map", &self.map)
+            .field("ptr", &self.ptr)
+            .field("name", &self.name)
+            .field("len", &self.block_len)
+            .field("type_ref", &self.type_ref)
+            .finish()
+    }
+}
+
+impl Eq for Branch {}
+
+impl PartialEq for Branch {
+    fn eq(&self, other: &Self) -> bool {
+        self.ptr == other.ptr
+            && self.start == other.start
+            && self.map == other.map
+            && self.name == other.name
+            && self.block_len == other.block_len
+            && self.type_ref == other.type_ref
+    }
 }
 
 impl Branch {
@@ -235,11 +279,12 @@ impl Branch {
         Self {
             start: None,
             map: HashMap::default(),
-            len: 0,
-            item: None,
+            block_len: 0,
+            content_len: 0,
             ptr,
             name,
             type_ref,
+            observers: None,
         }
     }
 
@@ -251,26 +296,30 @@ impl Branch {
     /// Returns a length of an indexed sequence component of a current branch node.
     /// Map component elements are computed on demand.
     pub fn len(&self) -> u32 {
-        self.len
+        self.block_len
+    }
+
+    pub fn content_len(&self, _: &Transaction) -> u32 {
+        self.content_len
     }
 
     /// Get iterator over (String, Block) entries of a map component of a current root type.
     /// Deleted blocks are skipped by this iterator.
-    pub(crate) fn entries<'a, 'b, 'txn>(&'a self, txn: &'b Transaction<'txn>) -> Entries<'b, 'txn> {
+    pub(crate) fn entries<'a, 'b>(&'a self, txn: &'b Transaction) -> Entries<'b> {
         Entries::new(&self.ptr, txn)
     }
 
     /// Get iterator over Block entries of an array component of a current root type.
     /// Deleted blocks are skipped by this iterator.
-    pub(crate) fn iter<'a, 'b, 'txn>(&'a self, txn: &'b Transaction<'txn>) -> Iter<'b, 'txn> {
+    pub(crate) fn iter<'a, 'b>(&'a self, txn: &'b Transaction) -> Iter<'b> {
         Iter::new(self.start, txn)
     }
 
     /// Returns a materialized value of non-deleted entry under a given `key` of a map component
     /// of a current root type.
-    pub(crate) fn get(&self, txn: &Transaction<'_>, key: &str) -> Option<Value> {
+    pub(crate) fn get(&self, txn: &Transaction, key: &str) -> Option<Value> {
         let ptr = self.map.get(key)?;
-        let item = txn.store.blocks.get_item(ptr)?;
+        let item = txn.store().blocks.get_item(ptr)?;
         if item.is_deleted() {
             None
         } else {
@@ -285,12 +334,12 @@ impl Branch {
     /// be returned.
     pub(crate) fn get_at<'a, 'b>(
         &'a self,
-        txn: &'b Transaction,
+        blocks: &'b BlockStore,
         mut index: u32,
     ) -> Option<(&'b ItemContent, usize)> {
         let mut ptr = self.start;
         while let Some(p) = ptr {
-            let item = txn.store.blocks.get_item(&p)?;
+            let item = blocks.get_item(&p)?;
             let len = item.len();
             if !item.is_deleted() && item.is_countable() {
                 if index < len {
@@ -310,7 +359,7 @@ impl Branch {
     pub(crate) fn remove(&self, txn: &mut Transaction, key: &str) -> Option<Value> {
         let ptr = self.map.get(key)?;
         let prev = {
-            let item = txn.store.blocks.get_item(ptr)?;
+            let item = txn.store().blocks.get_item(ptr)?;
             if item.is_deleted() {
                 None
             } else {
@@ -325,7 +374,7 @@ impl Branch {
     pub(crate) fn first<'a, 'b>(&'a self, txn: &'b Transaction) -> Option<&'b Item> {
         let mut ptr = self.start;
         while let Some(p) = ptr {
-            let item = txn.store.blocks.get_item(&p)?;
+            let item = txn.store().blocks.get_item(&p)?;
             if item.is_deleted() {
                 ptr = item.right.clone();
             } else {
@@ -352,37 +401,81 @@ impl Branch {
         mut ptr: Option<BlockPtr>,
         mut index: u32,
     ) -> (Option<BlockPtr>, Option<BlockPtr>) {
+        let store = txn.store_mut();
+        let encoding = store.options.offset_kind;
         while let Some(p) = ptr {
-            let item = txn
-                .store
+            let item = store
                 .blocks
                 .get_item(&p)
                 .expect("No item for a given pointer was found.");
-            let len = item.len();
+            let content_len = item.content_len(encoding);
             if !item.is_deleted() && item.is_countable() {
-                if index == len {
+                if index == content_len {
                     let left = Some(p.clone());
                     let right = item.right.clone();
                     return (left, right);
-                } else if index < len {
+                } else if index < content_len {
+                    let index = if let ItemContent::String(s) = &item.content {
+                        s.block_offset(index, encoding)
+                    } else {
+                        index
+                    };
                     let split_point = ID::new(item.id.client, item.id.clock + index);
                     let ptr = BlockPtr::new(split_point, p.pivot() as u32);
-                    let (left, mut right) = txn.store.blocks.split_block(&ptr);
+                    let (left, mut right) = store.blocks.split_block(&ptr);
                     if right.is_none() {
                         if let Some(left_ptr) = left.as_ref() {
-                            if let Some(left) = txn.store.blocks.get_item(left_ptr) {
+                            if let Some(left) = store.blocks.get_item(left_ptr) {
                                 right = left.right.clone();
                             }
                         }
                     }
                     return (left, right);
                 }
-                index -= len;
+                index -= content_len;
             }
             ptr = item.right.clone();
         }
         (None, None)
     }
+
+    pub(crate) fn path(from: Ref<Branch>, to: Ref<Branch>, txn: &Transaction) -> Path {
+        let parent = from;
+        let mut child = to;
+        let mut path = VecDeque::default();
+        while let TypePtr::Id(ptr) = &child.ptr {
+            if parent.ptr == child.ptr {
+                break;
+            }
+            let item = txn.store().blocks.get_item(ptr).unwrap();
+            if let Some(parent_sub) = item.parent_sub.clone() {
+                // parent is map-ish
+                path.push_front(PathSegment::Key(parent_sub));
+                child = txn.store().get_type(&item.parent).unwrap().borrow();
+            } else {
+                // parent is array-ish
+                let mut i = 0;
+                child = txn.store().get_type(&item.parent).unwrap().borrow();
+                let mut c = child.start.clone();
+                while let Some(ptr) = c {
+                    if ptr.id == item.id {
+                        break;
+                    }
+                    let cc = txn.store().blocks.get_block(&ptr).unwrap();
+                    if !cc.is_deleted() {
+                        i += 1;
+                    }
+                    if let Some(cci) = cc.as_item() {
+                        c = cci.right.clone();
+                    } else {
+                        break;
+                    }
+                }
+                path.push_front(PathSegment::Index(i));
+            }
+        }
+        path
+    }
 }
 
 /// Value that can be returned by Yrs data types. This includes [Any] which is an extension
@@ -398,6 +491,12 @@ pub enum Value {
     YXmlText(XmlText),
 }
 
+impl Default for Value {
+    fn default() -> Self {
+        Value::Any(Any::Null)
+    }
+}
+
 impl Value {
     /// Converts current value into [Any] object equivalent that resembles enhanced JSON payload.
     /// Rules are:
@@ -410,11 +509,11 @@ impl Value {
     pub fn to_json(self, txn: &Transaction) -> Any {
         match self {
             Value::Any(a) => a,
-            Value::YText(v) => Any::String(v.to_string(txn)),
+            Value::YText(v) => Any::String(v.to_string(txn).into_boxed_str()),
             Value::YArray(v) => v.to_json(txn),
             Value::YMap(v) => v.to_json(txn),
-            Value::YXmlElement(v) => Any::String(v.to_string(txn)),
-            Value::YXmlText(v) => Any::String(v.to_string(txn)),
+            Value::YXmlElement(v) => Any::String(v.to_string(txn).into_boxed_str()),
+            Value::YXmlText(v) => Any::String(v.to_string(txn).into_boxed_str()),
         }
     }
 
@@ -527,25 +626,25 @@ impl std::fmt::Display for Branch {
     }
 }
 
-pub(crate) struct Entries<'a, 'txn> {
-    pub txn: &'a Transaction<'txn>,
-    iter: std::collections::hash_map::Iter<'a, String, BlockPtr>,
+pub(crate) struct Entries<'a> {
+    pub txn: &'a Transaction,
+    iter: std::collections::hash_map::Iter<'a, Rc<str>, BlockPtr>,
 }
 
-impl<'a, 'txn> Entries<'a, 'txn> {
-    pub(crate) fn new<'b>(ptr: &'b TypePtr, txn: &'a Transaction<'txn>) -> Self {
-        let inner = txn.store.get_type(ptr).unwrap();
+impl<'a> Entries<'a> {
+    pub(crate) fn new<'b>(ptr: &'b TypePtr, txn: &'a Transaction) -> Self {
+        let inner = txn.store().get_type(ptr).unwrap();
         let iter = inner.as_ref().map.iter();
         Entries { txn, iter }
     }
 }
 
-impl<'a, 'txn> Iterator for Entries<'a, 'txn> {
-    type Item = (&'a String, &'a Item);
+impl<'a> Iterator for Entries<'a> {
+    type Item = (&'a str, &'a Item);
 
     fn next(&mut self) -> Option<Self::Item> {
         let (mut key, ptr) = self.iter.next()?;
-        let mut block = self.txn.store.blocks.get_item(ptr);
+        let mut block = self.txn.store().blocks.get_item(ptr);
         loop {
             match block {
                 Some(item) if !item.is_deleted() => {
@@ -554,7 +653,7 @@ impl<'a, 'txn> Iterator for Entries<'a, 'txn> {
                 _ => {
                     let (k, ptr) = self.iter.next()?;
                     key = k;
-                    block = self.txn.store.blocks.get_item(ptr);
+                    block = self.txn.store().blocks.get_item(ptr);
                 }
             }
         }
@@ -563,23 +662,23 @@ impl<'a, 'txn> Iterator for Entries<'a, 'txn> {
     }
 }
 
-pub(crate) struct Iter<'a, 'txn> {
+pub(crate) struct Iter<'a> {
     ptr: Option<BlockPtr>,
-    txn: &'a Transaction<'txn>,
+    txn: &'a Transaction,
 }
 
-impl<'a, 'txn> Iter<'a, 'txn> {
-    fn new(start: Option<BlockPtr>, txn: &'a Transaction<'txn>) -> Self {
+impl<'a> Iter<'a> {
+    fn new(start: Option<BlockPtr>, txn: &'a Transaction) -> Self {
         Iter { ptr: start, txn }
     }
 }
 
-impl<'a, 'txn> Iterator for Iter<'a, 'txn> {
+impl<'a> Iterator for Iter<'a> {
     type Item = &'a Item;
 
     fn next(&mut self) -> Option<Self::Item> {
         let ptr = self.ptr.take()?;
-        let item = self.txn.store.blocks.get_item(&ptr)?;
+        let item = self.txn.store().blocks.get_item(&ptr)?;
         self.ptr = item.right;
         Some(item)
     }
@@ -597,7 +696,7 @@ pub enum TypePtr {
     Id(block::BlockPtr),
 
     /// Pointer to a root-level type.
-    Named(Rc<String>),
+    Named(Rc<str>),
 }
 
 impl std::fmt::Display for TypePtr {
@@ -605,7 +704,258 @@ impl std::fmt::Display for TypePtr {
         match self {
             TypePtr::Unknown => write!(f, "unknown"),
             TypePtr::Id(ptr) => write!(f, "{}", ptr),
-            TypePtr::Named(name) => write!(f, "'{}'", name),
+            TypePtr::Named(name) => write!(f, "'{}'", name.as_ref()),
         }
     }
 }
+
+enum Observers {
+    Text(EventHandler<crate::types::text::TextEvent>),
+    Array(EventHandler<crate::types::array::ArrayEvent>),
+    Map(EventHandler<crate::types::map::MapEvent>),
+    Xml(EventHandler<crate::types::xml::XmlEvent>),
+    XmlText(EventHandler<crate::types::xml::XmlTextEvent>),
+}
+
+impl Observers {
+    pub fn text() -> Self {
+        Observers::Text(EventHandler::default())
+    }
+    pub fn array() -> Self {
+        Observers::Array(EventHandler::default())
+    }
+    pub fn map() -> Self {
+        Observers::Map(EventHandler::default())
+    }
+    pub fn xml() -> Self {
+        Observers::Xml(EventHandler::default())
+    }
+    pub fn xml_text() -> Self {
+        Observers::XmlText(EventHandler::default())
+    }
+
+    pub fn publish(
+        &self,
+        branch_ref: BranchRef,
+        txn: &Transaction,
+        keys: HashSet<Option<Rc<str>>>,
+    ) {
+        match self {
+            Observers::Text(eh) => eh.publish(txn, &TextEvent::new(branch_ref)),
+            Observers::Array(eh) => eh.publish(txn, &ArrayEvent::new(branch_ref)),
+            Observers::Map(eh) => eh.publish(txn, &MapEvent::new(branch_ref, keys)),
+            Observers::Xml(eh) => eh.publish(txn, &XmlEvent::new(branch_ref, keys)),
+            Observers::XmlText(eh) => eh.publish(txn, &XmlTextEvent::new(branch_ref, keys)),
+        }
+    }
+}
+
+/// A path describing nesting structure between shared collections containing each other. It's a
+/// collection of segments which refer to either index (in case of [Array] or [XmlElement]) or
+/// string key (in case of [Map]) where successor shared collection can be found within subsequent
+/// parent types.
+pub type Path = VecDeque<PathSegment>;
+
+/// A single segment of a [Path].
+pub enum PathSegment {
+    /// Key segments are used to inform how to access child shared collections within a [Map] types.
+    Key(Rc<str>),
+
+    /// Index segments are used to inform how to access child shared collections within an [Array]
+    /// or [XmlElement] types.
+    Index(u32),
+}
+
+pub(crate) struct ChangeSet<D> {
+    added: HashSet<ID>,
+    deleted: HashSet<ID>,
+    delta: Vec<D>,
+}
+
+impl<D> ChangeSet<D> {
+    pub fn new(added: HashSet<ID>, deleted: HashSet<ID>, delta: Vec<D>) -> Self {
+        ChangeSet {
+            added,
+            deleted,
+            delta,
+        }
+    }
+}
+
+/// A single change done over an array-component of shared data type.
+#[derive(Debug, Clone, PartialEq)]
+pub enum Change {
+    /// Determines a change that resulted in adding a consecutive number of new elements:
+    /// - For [Array] it's a range of inserted elements.
+    /// - For [XmlElement] it's a range of inserted child XML nodes.
+    Added(Vec<Value>),
+
+    /// Determines a change that resulted in removing a consecutive range of existing elements,
+    /// either XML child nodes for [XmlElement] or various elements stored in an [Array].
+    Removed(u32),
+
+    /// Determines a number of consecutive unchanged elements. Used to recognize non-edited spaces
+    /// between [Change::Added] and/or [Change::Removed] chunks.
+    Retain(u32),
+}
+
+/// A single change done over a map-component of shared data type.
+#[derive(Debug, Clone, PartialEq)]
+pub enum EntryChange {
+    /// Informs about a new value inserted under specified entry.
+    Inserted(Value),
+
+    /// Informs about a change of old value (1st field) to a new one (2nd field) under
+    /// a corresponding entry.
+    Updated(Value, Value),
+
+    /// Informs about a removal of a corresponding entry - contains a removed value.
+    Removed(Value),
+}
+
+/// A single change done over a text-like types: [Text] or [XmlText].
+#[derive(Debug, Clone, PartialEq)]
+pub enum Delta {
+    /// Determines a change that resulted in insertion of a piece of text, which optionally could
+    /// have been formatted with provided set of attributes.
+    Inserted(Value, Option<Box<Attrs>>),
+
+    /// Determines a change that resulted in removing a consecutive range of characters.
+    Deleted(u32),
+
+    /// Determines a number of consecutive unchanged characters. Used to recognize non-edited spaces
+    /// between [Delta::Inserted] and/or [Delta::Deleted] chunks. Can contain an optional set of
+    /// attributes, which have been used to format an existing piece of text.
+    Retain(u32, Option<Box<Attrs>>),
+}
+
+/// An alias for map of attributes used as formatting parameters by [Text] and [XmlText] types.
+pub type Attrs = HashMap<Box<str>, Any>;
+
+pub(crate) fn event_keys(
+    txn: &Transaction,
+    target: Ref<Branch>,
+    keys_changed: &HashSet<Option<Rc<str>>>,
+) -> HashMap<Rc<str>, EntryChange> {
+    let mut keys = HashMap::new();
+    for opt in keys_changed.iter() {
+        if let Some(key) = opt {
+            let item = target
+                .map
+                .get(key.as_ref())
+                .and_then(|ptr| txn.store().blocks.get_item(ptr));
+            if let Some(item) = item {
+                if item.id.clock >= txn.before_state.get(&item.id.client) {
+                    let mut prev = item
+                        .left
+                        .as_ref()
+                        .and_then(|ptr| txn.store().blocks.get_item(ptr));
+                    while let Some(p) = prev {
+                        if !txn.has_added(&p.id) {
+                            break;
+                        }
+                        prev = p
+                            .left
+                            .as_ref()
+                            .and_then(|ptr| txn.store().blocks.get_item(ptr));
+                    }
+
+                    if txn.has_deleted(&item.id) {
+                        if let Some(prev) = prev {
+                            if txn.has_deleted(&prev.id) {
+                                let old_value =
+                                    prev.content.get_content_last(txn).unwrap_or_default();
+                                keys.insert(key.clone(), EntryChange::Removed(old_value));
+                            }
+                        }
+                    } else {
+                        let new_value = item.content.get_content_last(txn).unwrap();
+                        if let Some(prev) = prev {
+                            if txn.has_deleted(&prev.id) {
+                                let old_value =
+                                    prev.content.get_content_last(txn).unwrap_or_default();
+                                keys.insert(
+                                    key.clone(),
+                                    EntryChange::Updated(old_value, new_value),
+                                );
+
+                                continue;
+                            }
+                        }
+
+                        keys.insert(key.clone(), EntryChange::Inserted(new_value));
+                    }
+                } else if txn.has_deleted(&item.id) {
+                    let old_value = item.content.get_content_last(txn).unwrap_or_default();
+                    keys.insert(key.clone(), EntryChange::Removed(old_value));
+                }
+            }
+        }
+    }
+
+    keys
+}
+
+pub(crate) fn event_change_set(txn: &Transaction, start: Option<&BlockPtr>) -> ChangeSet<Change> {
+    let mut added = HashSet::new();
+    let mut deleted = HashSet::new();
+    let mut delta = Vec::new();
+
+    let mut last_op = None;
+
+    let mut current = start.and_then(|ptr| txn.store().blocks.get_item(&ptr));
+
+    while let Some(item) = current {
+        if item.is_deleted() {
+            if txn.has_deleted(&item.id) && !txn.has_added(&item.id) {
+                let removed = match last_op.take() {
+                    None => 0,
+                    Some(Change::Removed(c)) => c,
+                    Some(other) => {
+                        delta.push(other);
+                        0
+                    }
+                };
+                last_op = Some(Change::Removed(removed + item.len()));
+                deleted.insert(item.id);
+            } // else nop
+        } else {
+            if txn.has_added(&item.id) {
+                let mut inserts = match last_op.take() {
+                    None => Vec::with_capacity(item.len() as usize),
+                    Some(Change::Added(values)) => values,
+                    Some(other) => {
+                        delta.push(other);
+                        Vec::with_capacity(item.len() as usize)
+                    }
+                };
+                inserts.append(&mut item.content.get_content(txn));
+                last_op = Some(Change::Added(inserts));
+                added.insert(item.id);
+            } else {
+                let retain = match last_op.take() {
+                    None => 0,
+                    Some(Change::Retain(c)) => c,
+                    Some(other) => {
+                        delta.push(other);
+                        0
+                    }
+                };
+                last_op = Some(Change::Retain(retain + item.len()));
+            }
+        }
+
+        current = item
+            .right
+            .as_ref()
+            .and_then(|ptr| txn.store().blocks.get_item(ptr));
+    }
+
+    match last_op.take() {
+        Some(Change::Retain(_)) => { /* do nothing */ }
+        Some(change) => delta.push(change),
+        None => { /* do nothing */ }
+    }
+
+    ChangeSet::new(added, deleted, delta)
+}
